/*
 * Copyright (c) 2022, Peter Abeles. All Rights Reserved.
 *
 * This file is part of BoofCV (http://boofcv.org).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package boofcv.struct.image;

import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Array;

/**
 * <p>
 * Multi-band image composed of discontinuous planar images for each band. The bands are discontinuous in that
 * each one is an independent memory and are of type {@link ImageGray}. Planar images fully supports all
 * functions inside of {@link ImageBase}. Each internal image has the same width, height, startIndex, and stride.
 * </p>
 *
 * <p>
 * For example, in a RGB image there would be three bands one for each color,
 * and each color would be stored in its own gray scale image. To access the image for a particular
 * band call {@link #getBand(int)}. To get the RGB value for a pixel (x,y) one would need to:
 * <pre>
 * int red = image.getBand(0).get(x,y);
 * int green = image.getBand(1).get(x,y);
 * int blue = image.getBand(2).get(x,y);
 * </pre>
 * Setting the RGB value of pixel (x,y) is done in a similar manner:
 * <pre>
 * image.getBand(0).set(x,y,red);
 * image.getBand(1).set(x,y,green);
 * image.getBand(2).set(x,y,blue);
 * </pre>
 * </p>
 *
 * <p>
 * May image processing operations can be run independently on each color band. This is useful since many
 * operations have been written for {@link ImageGray}, but not Planar yet.
 * <pre>
 * for( int i = 0; i < image.numBands(); i++ ) {
 *     SomeGrayImageFilter.process( image.getBand(0) );
 * }
 * </pre>
 * </p>
 *
 * @author Peter Abeles
 */
public class Planar<T extends ImageGray<T>> extends ImageMultiBand<Planar<T>> {

	/**
	 * Type of image in each band
	 */
	public Class<T> type;

	/**
	 * Set of gray scale images
	 */
	public T[] bands;

	/**
	 * Creates a Planar image with the specified properties.
	 *
	 * @param type The type of image which each band is stored as.
	 * @param width Width of the image.
	 * @param height Height of the image.
	 * @param numBands Total number of bands.
	 */
	public Planar( Class<T> type, int width, int height, int numBands ) {
		this.type = type;
		this.stride = width;
		this.width = width;
		this.height = height;
		this.bands = (T[])Array.newInstance(type, numBands);

		for (int i = 0; i < numBands; i++) {
			bands[i] = ImageGray.create(type, width, height);
		}
		this.imageType = ImageType.pl(numBands, type);
	}

	/**
	 * Declares internal arrays for storing each band, but not the images in each band.
	 *
	 * @param type The type of image which each band is stored as.
	 * @param numBands Number of bands in the image.
	 */
	public Planar( Class<T> type, int numBands ) {
		this.type = type;
		this.bands = (T[])Array.newInstance(type, numBands);
		this.imageType = ImageType.pl(numBands, type);
	}

	/**
	 * Type of image each band is stored as.
	 *
	 * @return The type of ImageGray which each band is stored as.
	 */
	public Class<T> getBandType() {
		return type;
	}

	/**
	 * Returns the number of bands or colors stored in this image.
	 *
	 * @return Number of bands in the image.
	 */
	@Override
	public int getNumBands() {
		return bands.length;
	}

	/**
	 * Returns a band in the multi-band image.
	 *
	 * @param band Which band should be returned.
	 * @return Image band
	 */
	public T getBand( int band ) {
		if (band >= bands.length || band < 0)
			throw new IllegalArgumentException("The specified band is out of range. " + band + " / " + bands.length);

		return bands[band];
	}

	/**
	 * Creates a sub-image from 'this' image. The subimage will share the same internal array
	 * that stores each pixel's value, but will only pertain to an axis-aligned rectangular segment
	 * of the original.
	 *
	 * @param x0 x-coordinate of top-left corner of the sub-image.
	 * @param y0 y-coordinate of top-left corner of the sub-image.
	 * @param x1 x-coordinate of bottom-right corner of the sub-image.
	 * @param y1 y-coordinate of bottom-right corner of the sub-image.
	 * @param output Optional storage for subimage
	 * @return A sub-image of this image.
	 */
	@Override
	public Planar<T> subimage( int x0, int y0, int x1, int y1, @Nullable Planar<T> output ) {
		if (x0 < 0 || y0 < 0)
			throw new IllegalArgumentException("x0 or y0 is less than zero");
		if (x1 < x0 || y1 < y0)
			throw new IllegalArgumentException("x1 or y1 is less than x0 or y0 respectively");
		if (x1 > width || y1 > height)
			throw new IllegalArgumentException("x1 or y1 is more than the width or height respectively");

		if (output != null) {
			output.type = type;
			output.bands = (T[])Array.newInstance(type, bands.length);
		} else {
			output = new Planar<>(type, bands.length);
		}

		output.stride = Math.max(width, stride);
		output.width = x1 - x0;
		output.height = y1 - y0;
		output.startIndex = startIndex + y0*stride + x0;
		output.subImage = true;

		for (int i = 0; i < bands.length; i++) {
			output.bands[i] = (T)bands[i].subimage(x0, y0, x1, y1);
		}

		return output;
	}

	/**
	 * Sets the values of each pixel equal to the pixels in the specified matrix.
	 * Automatically resized to match the input image.
	 *
	 * @param orig The original image whose value is to be copied into this one
	 */
	@Override
	public Planar<T> setTo( Planar<T> orig ) {
		if (orig.width != width || orig.height != height)
			reshape(orig.width, orig.height);
		if (orig.getBandType() != getBandType())
			throw new IllegalArgumentException("The band type must be the same");

		int N = orig.getNumBands();
		if (N != getNumBands()) {
			setNumberOfBands(orig.getNumBands());
		}

		for (int i = 0; i < N; i++) {
			bands[i].setTo(orig.getBand(i));
		}
		return this;
	}

	/**
	 * Changes the image's width and height without declaring new memory. If the internal array
	 * is not large enough to store the new image an IllegalArgumentException is thrown.
	 *
	 * @param width The new width.
	 * @param height The new height.
	 */
	@Override
	public void reshape( int width, int height ) {
		if (this.width == width && this.height == height)
			return;

		if (isSubimage())
			throw new IllegalArgumentException("Can't reshape subimage");

		for (int i = 0; i < bands.length; i++) {
			bands[i].reshape(width, height);
		}

		this.startIndex = 0;
		this.stride = width;
		this.width = width;
		this.height = height;
	}

	@Override
	public void reshape( int width, int height, int numberOfBands ) {
		if (getNumBands() != numberOfBands) {
			if (isSubimage())
				throw new RuntimeException("Can't reshape subimage");

			T[] bands = (T[])Array.newInstance(type, numberOfBands);
			int N = Math.min(numberOfBands, this.bands.length);
			for (int i = 0; i < N; i++) {
				bands[i] = this.bands[i];
				bands[i].reshape(width, height);
			}
			for (int i = N; i < bands.length; i++) {
				bands[i] = ImageGray.create(type, width, height);
			}
			this.startIndex = 0;
			this.bands = bands;
			this.stride = width;
			this.width = width;
			this.height = height;
			this.imageType.numBands = numberOfBands;
		} else {
			reshape(width, height);
		}
	}

	/**
	 * Creates a new image of the same type and number of bands
	 *
	 * @param imgWidth image width
	 * @param imgHeight image height
	 * @return new image
	 */
	@Override
	public Planar<T> createNew( int imgWidth, int imgHeight ) {
		return new Planar<>(type, imgWidth, imgHeight, bands.length);
	}

	@Override
	public void copyRow( int row, int col0, int col1, int offset, Object array ) {
		throw new IllegalArgumentException("Not supported for planar images");
	}

	@Override
	public void copyCol( int col, int row0, int row1, int offset, Object array ) {
		throw new IllegalArgumentException("Not supported for planar images");
	}

	/**
	 * Returns a new {@link Planar} which references the same internal single band images at this one.
	 *
	 * @param which List of bands which will comprise the new image
	 * @return New image
	 */
	public Planar<T> partialSpectrum( int... which ) {
		Planar<T> out = new Planar<>(getBandType(), which.length);

		out.setWidth(width);
		out.setHeight(height);
		out.setStride(stride);

		for (int i = 0; i < which.length; i++) {
			out.setBand(i, getBand(which[i]));
		}

		return out;
	}

	/**
	 * Changes the bands order
	 *
	 * @param order The new band order
	 */
	public void reorderBands( int... order ) {
		T[] bands = (T[])Array.newInstance(type, order.length);

		for (int i = 0; i < order.length; i++) {
			bands[i] = this.bands[order[i]];
		}
		this.bands = bands;
	}

	/**
	 * Changes the number of bands in the image. A new array is declared and individual bands are recycled
	 * if possible
	 *
	 * @param numberOfBands New number of bands in the image.
	 */
	@Override
	public void setNumberOfBands( int numberOfBands ) {
		if (numberOfBands == this.bands.length)
			return;

		T[] bands = (T[])Array.newInstance(type, numberOfBands);
		int N = Math.min(numberOfBands, this.bands.length);
		for (int i = 0; i < N; i++) {
			bands[i] = this.bands[i];
		}
		for (int i = N; i < bands.length; i++) {
			bands[i] = ImageGray.create(type, width, height);
		}
		this.bands = bands;
		this.imageType.numBands = this.bands.length;
	}

	/**
	 * Returns an integer formed from 4 bands. band[0]<<24 | band[1] << 16 | band[2]<<8 | band[3]. Assumes
	 * arrays are U8 type.
	 *
	 * @param x column
	 * @param y row
	 * @return 32 bit integer
	 */
	public int get32u8( int x, int y ) {
		int i = startIndex + y*stride + x;
		return ((((GrayU8)bands[0]).data[i] & 0xFF) << 24) |
				((((GrayU8)bands[1]).data[i] & 0xFF) << 16) |
				((((GrayU8)bands[2]).data[i] & 0xFF) << 8) |
				(((GrayU8)bands[3]).data[i] & 0xFF);
	}

	/**
	 * Returns an integer formed from 3 bands. band[0]<<16| band[1] << 8 | band[2]
	 *
	 * @param x column
	 * @param y row
	 * @return 32 bit integer
	 */
	public int get24u8( int x, int y ) {
		int i = startIndex + y*stride + x;
		return ((((GrayU8)bands[0]).data[i] & 0xFF) << 16) |
				((((GrayU8)bands[1]).data[i] & 0xFF) << 8) |
				(((GrayU8)bands[2]).data[i] & 0xFF);
	}

	public void set24u8( int x, int y, int value ) {
		int i = startIndex + y*stride + x;
		((GrayU8)bands[0]).data[i] = (byte)(value >>> 16);
		((GrayU8)bands[1]).data[i] = (byte)(value >>> 8);
		((GrayU8)bands[2]).data[i] = (byte)value;
	}

	public void set32u8( int x, int y, int value ) {
		int i = startIndex + y*stride + x;
		((GrayU8)bands[0]).data[i] = (byte)(value >>> 24);
		((GrayU8)bands[1]).data[i] = (byte)(value >>> 16);
		((GrayU8)bands[2]).data[i] = (byte)(value >>> 8);
		((GrayU8)bands[3]).data[i] = (byte)value;
	}

	public void setBandType( Class<T> type ) {
		if (this.type != null && this.type != type)
			throw new RuntimeException("Once the band type has been set you can't change it");
		this.type = type;
	}

	public T[] getBands() {
		return bands;
	}

	public void setBands( T[] bands ) {
		this.bands = bands;
	}

	public void setBand( int which, T image ) {
		this.bands[which] = image;
	}
}
